'''
Created on 2021-07-39

@author: wf
'''
from corpus.event import EventStorage
from corpus.eventcorpus import EventCorpus, EventDataSource

from datasources.dblp import DblpEventManager,DblpEventSeriesManager
from datasources.wikidata import Wikidata,WikidataEventManager,WikidataEventSeriesManager
from datasources.openresearch import OREventManager,OREventSeriesManager
from datasources.wikicfp import WikiCfpEventManager,WikiCfpEventSeriesManager
from datasources.crossref import CrossrefEventManager,CrossrefEventSeriesManager
from lodstorage.uml import UML

from datetime import datetime

import os
import sys

from argparse import ArgumentParser
from argparse import RawDescriptionHelpFormatter

class CorpusLookup(object):
    '''
    search and lookup for different EventCorpora
    '''
    lookupIds=["dblp","crossref","wikidata","wikicfp","or","or-backup","orclone","orclone-backup"]

    def __init__(self,lookupIds:list=None,
                 configure:callable=None,debug=False):
        '''
        Constructor
        
        Args:
            lookupIds(list): the list of lookupIds to addDataSources for
            configure(callable): Callback to configure the corpus lookup
        '''
        self.debug=debug
        self.configure=configure
        self.eventCorpus=EventCorpus()
        if lookupIds is None:
            lookupIds=CorpusLookup.lookupIds
        if "crossref" in lookupIds:
            self.eventCorpus.addDataSource(CrossrefEventManager(),CrossrefEventSeriesManager(),lookupId="crossref",name="crossref.org",url="https://www.crossref.org/",title="CrossRef",tablePrefix="crossref")
        if "dblp" in lookupIds:
            self.eventCorpus.addDataSource(DblpEventManager(),DblpEventSeriesManager(),lookupId="dblp",name="dblp",url='https://dblp.org/',title='dblp computer science bibliography',tablePrefix="dblp")
        if "wikidata" in lookupIds: 
            self.eventCorpus.addDataSource(WikidataEventManager(),WikidataEventSeriesManager(),lookupId="wikidata",name="Wikidata",url='https://www.wikidata.org/wiki/Wikidata:Main_Page',title='Wikidata',tablePrefix="wikidata")
        if "wikicfp" in lookupIds:    
            self.eventCorpus.addDataSource(WikiCfpEventManager(),WikiCfpEventSeriesManager(),lookupId="wikicfp",name="WikiCFP",url='http://www.wikicfp.com',title='WikiCFP',tablePrefix="wikicfp")
        if "or" in lookupIds:    
            self.eventCorpus.addDataSource(OREventManager(),OREventSeriesManager(),lookupId="or",name="OR_Triples",url='https://www.openresearch.org/wiki/Main_Page',title='OPENRESEARCH-api',tablePrefix="orapi")
        if "or-backup" in lookupIds:    
            self.eventCorpus.addDataSource(OREventManager(),OREventSeriesManager(),lookupId="or-backup",name="OR_Markup",url='https://www.openresearch.org/wiki/Main_Page',title='OPENRESEARCH-wiki',tablePrefix="orwiki")
        if "orclone" in lookupIds:    
            self.eventCorpus.addDataSource(OREventManager(),OREventSeriesManager(),lookupId="orclone",name="OR_Clone_Triples",url='https://confident.dbis.rwth-aachen.de/or/index.php?title=Main_Page',title='OPENRESEARCH-clone-api',tablePrefix="orcapi")
        if "orclone-backup" in lookupIds:    
            self.eventCorpus.addDataSource(OREventManager(),OREventSeriesManager(),lookupId="orclone-backup",name="OR_Clone_Markup",url='https://confident.dbis.rwth-aachen.de/or/index.php?title=Main_Page',title='OPENRESEARCH-clone-wiki',tablePrefix="orcwiki")
        
    def getDataSource(self,lookupId:str)->EventDataSource:
        '''
        get the given data source
        
        Args:
            lookupId(str): the lookupId of the data source to get
            
        Return:
            EventDataSource: the data source

        '''
        eventDataSource=None
        if lookupId in self.eventCorpus.eventDataSources:
            eventDataSource=self.eventCorpus.eventDataSources[lookupId]
        return eventDataSource
        
    def load(self,forceUpdate:bool=False):
        '''
        load the event corpora
        Args:
            forceUpdate(bool): True if the data should be fetched from the source instead of the cache
        '''
        if self.configure:
            self.configure(self)
        self.eventCorpus.loadAll(forceUpdate=forceUpdate)
        
    def asPlantUml(self,baseEntity='Event'):
        '''
        return me as a plantUml Diagram markup
        '''
        storageTableList=EventStorage.getTableList()
        schemaManager=None
        uml=UML()
        now=datetime.now()
        nowYMD=now.strftime("%Y-%m-%d")
        tableList=[]
        for table in storageTableList:
            tableName=table['name']
            if tableName.endswith(baseEntity):
                if 'instances' in table:
                    instanceNote=""
                    instanceCount=table['instances']
                    instanceNote=f"\n{instanceCount} instances "
                    table['notes']=instanceNote
                tableList.append(table)
        title=f"""ConfIDent  {baseEntity}
{nowYMD}
[[https://projects.tib.eu/en/confident/ © 2019-2021 ConfIDent project]]
see also [[http://ptp.bitplan.com/settings Proceedings Title Parser]]
"""
        plantUml=uml.mergeSchema(schemaManager,tableList,title=title,packageName='DataSources',generalizeTo=baseEntity)
        return plantUml
        
        
__version__ = "0.0.7"
__date__ = '2020-06-22'
__updated__ = '2021-08-02'    

DEBUG = 1

    
def main(argv=None): # IGNORE:C0111
    '''main program.'''

    if argv is None:
        argv = sys.argv
    else:
        sys.argv.extend(argv)    
        
    program_name = os.path.basename(sys.argv[0])
    program_version = "v%s" % __version__
    program_build_date = str(__updated__)
    program_version_message = '%%(prog)s %s (%s)' % (program_version, program_build_date)
    program_shortdesc = __import__('__main__').__doc__.split("\n")[1]
    user_name="Wolfgang Fahl"
    program_license = '''%s

  Created by %s on %s.
  Copyright 2020-2021 Wolfgang Fahl. All rights reserved.

  Licensed under the Apache License 2.0
  http://www.apache.org/licenses/LICENSE-2.0

  Distributed on an "AS IS" basis without warranties
  or conditions of any kind, either express or implied.

USAGE
''' % (program_shortdesc, user_name,str(__date__))

    try:
        # Setup argument parser
        datasourcesDefault=",".join(CorpusLookup.lookupIds)
        parser = ArgumentParser(description=program_license, formatter_class=RawDescriptionHelpFormatter)
        parser.add_argument("-d", "--debug", dest="debug", action="store_true", help="show debug info")
        parser.add_argument('-e', '--endpoint', default=Wikidata.endpoint, help="SPARQL endpoint to use for wikidata queries")     
        parser.add_argument('-v', '--version', action='version', version=program_version_message)
        parser.add_argument("-u", "--uml", dest="uml", action="store_true", help="output plantuml diagram markup")
        parser.add_argument("-f", "--force",dest="forceUpdate",action="store_true",help="force Update - may take quite a time")
        parser.add_argument("--datasources",help=", delimited list of datasource lookup ids",default=datasourcesDefault)
        
        # Process arguments
        args = parser.parse_args()   
        Wikidata.endpoint=args.endpoint
        lookupIds=args.datasources.split(",")
        lookup=CorpusLookup(debug=args.debug,lookupIds=lookupIds)
        lookup.load(forceUpdate=args.forceUpdate)
        if args.uml:
            for baseEntity in ["Event","EventSeries"]:
                plantUml=lookup.asPlantUml(baseEntity)
            print(plantUml)
        
        
    except KeyboardInterrupt:
        ### handle keyboard interrupt ###
        return 1
    except Exception as e:
        if DEBUG:
            raise(e)
        indent = len(program_name) * " "
        sys.stderr.write(program_name + ": " + repr(e) + "\n")
        sys.stderr.write(indent + "  for help use --help")
        return 2         
        
if __name__ == "__main__":
    if DEBUG:
        sys.argv.append("-d")
    sys.exit(main())



'''
Created on 2021-07-26

@author: wf
'''
from lodstorage.entity import EntityManager
from lodstorage.jsonable import JSONAble
from lodstorage.lod import LOD
from lodstorage.sql import SQLDB
from lodstorage.storageconfig import StorageConfig

class EventStorage:
    '''
    common storage aspects of the EventManager and EventSeriesManager
    '''
    @staticmethod
    def getStorageConfig(debug:bool=False,mode='sql')->StorageConfig:
        '''
        get the storageConfiguration
        
        Args:
            debug(bool): if True show debug information
            mode(str): sql or json
        
        Return:
            StorageConfig: the storage configuration to be used
        '''
        if mode=='sql':
            config=StorageConfig.getSQL(debug=debug)
        elif mode=='json':
            config=StorageConfig.getJSON()
        elif mode=='jsonpickle':
            config=StorageConfig.getJsonPickle(debug=debug)
        else:
            raise Exception(f"invalid mode {mode}")
        config.cacheDirName="conferencecorpus"
        cachedir=config.getCachePath() 
        if mode=='sql':
            config.cacheFile=f"{cachedir}/EventCorpus.db"
        return config
    
    @staticmethod
    def getTableList()->list:
        '''
        get the list of SQL Tables involved
        
        Return:
            list: the map of SQL tables used for caching
        '''
        config=EventStorage.getStorageConfig()
        
        sqlDB=SQLDB(config.cacheFile)
        tableList=sqlDB.getTableList()
        for table in tableList:
            tableName=table["name"]
            countQuery="SELECT count(*) as count from %s" % tableName
            countResult=sqlDB.query(countQuery)
            table['instances']=countResult[0]['count']
        return tableList
    

class Event(JSONAble):
    '''
    base class for Event entities
    '''
    def __init__(self):
        '''
        Constructor
        '''

    def __str__(self):
        '''
        return my
        '''
        text=self.__class__.__name__
        attrs=["pageTitle","acronym","eventId","title","year","source","url"]
        delim=":"
        for attr in attrs:
            if hasattr(self, attr):
                value=getattr(self,attr)
                text+=f"{delim}{value}"
                delim=":" 
        return text
    
class EventSeries(JSONAble):
    '''
    base class for Event Series entities
    '''
    def __init__(self):
        '''
        Constructor
        '''
        pass

class EventBaseManager(EntityManager):
    '''
    common entity Manager for ConferenceCorpus
    '''
    
    def __init__(self,name,entityName,entityPluralName:str,listName:str=None,clazz=None,tableName:str=None,primaryKey:str=None,config=None,handleInvalidListTypes=False,filterInvalidListTypes=False,debug=False,profile=True):
        '''
        Constructor
        
        Args:
            name(string): name of this eventManager
            entityName(string): entityType to be managed e.g. Country
            entityPluralName(string): plural of the the entityType e.g. Countries
            config(StorageConfig): the configuration to be used if None a default configuration will be used
            handleInvalidListTypes(bool): True if invalidListTypes should be converted or filtered
            filterInvalidListTypes(bool): True if invalidListTypes should be deleted
            debug(boolean): override debug setting when default of config is used via config=None
            profile(boolean): True if profiling/timing information should be shown for long-running operations
        '''
        self.profile=profile
        if config is None:
            config=EventStorage.getStorageConfig(debug=debug)
        super().__init__(name, entityName, entityPluralName, listName, clazz, tableName, primaryKey, config, handleInvalidListTypes, filterInvalidListTypes, debug)
        
    def configure(self):
        '''
        configure me - abstract method that needs to be overridden
        '''    
        raise Exception(f"specialized configure for {self.name} needs to be implemented")
 
    def setAllAttr(self,listOfDicts,attr,value):
        '''
        set all attribute values of the given attr in the given list of Dict to the given value 
        '''
        for record in listOfDicts:
            record[attr]=value
    
class EventSeriesManager(EventBaseManager):
    '''
    Event series list
    '''
    def __init__(self,name:str,clazz=None,tableName:str=None,primaryKey:str=None,config:StorageConfig=None,debug=False):
        '''
        constructor 
        '''
        super().__init__(name=name,entityName="EventSeries",entityPluralName="EventSeries",primaryKey=primaryKey,listName="series",clazz=clazz,tableName=tableName,handleInvalidListTypes=True,config=config,debug=debug)
        
            
class EventManager(EventBaseManager):
    '''
    Event entity list
    '''
    
    def __init__(self,name:str,clazz=None,tableName:str=None,primaryKey:str=None,config:StorageConfig=None,debug=False):
        '''
        constructor 
        '''
        super(EventManager, self).__init__(name=name,entityName="Event",entityPluralName="Events",primaryKey=primaryKey,listName="events",clazz=clazz,tableName=tableName,config=config,handleInvalidListTypes=True,debug=debug)
        
 
    def linkSeriesAndEvent(self, eventSeriesManager:EventSeriesManager, seriesKey:str="series"):
        '''
        link Series and Event using the given foreignKey

        Args:
            seriesKey(str): the key to be use for lookup
            eventSeriesManager(EventSeriesManager):
        '''
        # get foreign key hashtable
        self.seriesLookup = LOD.getLookup(self.getList(), seriesKey, withDuplicates=True)
        # get "primary" key hashtable
        self.seriesAcronymLookup = LOD.getLookup(eventSeriesManager.getList(), "acronym", withDuplicates=True)

        for seriesAcronym in self.seriesLookup.keys():
            if seriesAcronym in self.seriesAcronymLookup:
                seriesEvents = self.seriesLookup[seriesAcronym]
                if self.verbose:
                    print(f"{seriesAcronym}:{len(seriesEvents):4d}")
            else:
                if self.debug:
                    print(f"Event Series Acronym {seriesAcronym} lookup failed")
        if self.debug:
            print("%d events/%d eventSeries -> %d linked" % (
            len(self.getList()), len(eventSeriesManager.getList()), len(self.seriesLookup)))
            
    def getEventsInSeries(self,seriesAcronym):
        """
        Return all the events in a given series.
        """
        if seriesAcronym in self.seriesAcronymLookup:
            seriesEvents = self.seriesLookup[seriesAcronym]
            if self.debug:
                print(f"{seriesAcronym}:{len(seriesEvents):4d}")
        else:
            if self.debug:
                print(f"Event Series Acronym {seriesAcronym} lookup failed")
            return None
        return seriesEvents



'''
Created on 2020-07-05

@author: wf
'''
import habanero
from corpus.event import Event,EventSeries,EventManager,EventSeriesManager
from lodstorage.storageconfig import StorageConfig
from lodstorage.lod import LOD
import json
import re
import glob
import os
import time
import datetime


class Crossref(object):
    '''
        Access to Crossref's search api see https://github.com/CrossRef/rest-api-doc
    '''
    def __init__(self):
        '''
        constructor
        '''
        self.cr=habanero.Crossref()   
     
    def doiMetaData(self,doi):
        ''' get the meta data for the given doi '''
        metadata=None
        response=self.cr.works([doi])
        if 'status' in response and 'message' in response and response['status']=='ok':
            metadata=response['message']
        return metadata
    
    @staticmethod    
    def fixEncodings(eventInfo:dict,debug:bool=False):    
        '''
        fix the encoding of the dict entries of the given eventInfo
        
        Args:
            eventInfo(dict): the record to fix
            debug(bool): show debug information if True
        '''
        for keyValue in eventInfo.items():
            key,value=keyValue
            oldvalue=value
            if isinstance(value,str):
                # work around Umlaut encodings like "M\\"unster"
                # and \S encoded as \\S
                found=False
                # see also https://github.com/WolfgangFahl/ProceedingsTitleParser/issues/38
                # remove encoded CR 
                for umlautTuple in [('\\"a',"ä"),('\\"o',"ö"),('\\"u',"ü"),('\\',' '),('&#x0D;','')]:
                    uc,u=umlautTuple
                    if uc in value:
                        value=value.replace(uc,u)
                        found=True  
                if found:
                    if debug:
                        print("Warning: fixing '%s' to '%s'" % (oldvalue,value))
                    eventInfo[key]=value   
    
class CrossrefEvent(Event):
    '''
    a scientific event derived from Crossref
    '''
    
class CrossrefEventSeries(Event):
    '''
    a scientific event series derived fro Crossref
    '''
      

class CrossrefEventManager(EventManager):
    '''
    Crossref event manager
    '''
        
    def __init__(self, config: StorageConfig = None):
        '''
        Constructor
        '''
        super().__init__(name="CrossrefEvents", clazz=CrossrefEvent,
                                                         tableName="crossref_event", config=config)
        
    def configure(self):
        '''
        configure me
        '''
        # nothing to do - there is a get ListOfDicts below    

        
    def getListOfDicts(self):
        '''
        cache my json files to my eventmanager
        '''
        startTime=time.time()
        jsonFiles=self.jsonFiles()
        lod=[]
        if len(jsonFiles)<50:
            print ("not enough crossref json files - did getsamples fail due to crossref API issues?")
            print ("need to download cached version of json files as a work-around instead")#
            raise Exception(f"only {len(jsonFiles)} crossref json files available - there should be at least 47")
            #url="http://wiki.bitplan.com/images/confident/Event_crossref.db"
            #filename=self.em.getCacheFile(mode=StoreMode.SQL)
            #urllib.request.urlretrieve(url, filename)
        for jsonFilePath in jsonFiles:
            eventBatch=self.fromJsonFile(jsonFilePath)
            if self.debug:
                print("%4d: %s" % (len(eventBatch),jsonFilePath))
            for eventInfo in eventBatch:
                rawEvent=self.postProcess(eventInfo)
                lod.append(rawEvent)
                    
        if self.profile:
            elapsed=time.time()-startTime
            print (f"read {len(lod)} events in {elapsed:5.1f} s" )
        return lod
            
        
    def postProcess(self,eventInfo:dict)->dict:
        '''
        postProcess the given eventInfo e.g.
         {
        "event": {
          "name": "Adriatico Research Conference and Workshop",
          "location": "ICTP, Trieste, Italy"
        },
        "title": [
          "Towards the Theoretical Understanding of High Tc Superconductors"
        ],
        "DOI": "10.1142/0638"
      },

        '''
        rawEvent=eventInfo['event']
        rawEvent['source']="crossref"
        if 'title' in eventInfo:
            title=eventInfo["title"][0]
            rawEvent['title']=title
        if 'sponsor' in eventInfo:
            sponsor=eventInfo['sponsor'][0]
            rawEvent['sponsor']=sponsor    
        Crossref.fixEncodings(rawEvent,self.debug)
                            
        doi=eventInfo["DOI"]
        rawEvent['eventId']=doi
        rawEvent['doi']=doi
        if 'start' in rawEvent: 
            self.fixDateParts(rawEvent,'start')
        if 'end'   in rawEvent: 
            self.fixDateParts(rawEvent,'end')
        LOD.setNone(rawEvent, ['lookupAcronym','location','number','sponsor','month','year','startDate','endDate'])
        if 'year' in eventInfo:
            year=eventInfo["year"]
            if year is not None and type(year) is tuple:
                year=year[0]
                if not year in rawEvent:
                    rawEvent["year"]=int(year)
        rawEvent["url"]=f"https://api.crossref.org/v1/works/{doi}" 
        return rawEvent
        
    def fixDateParts(self,rawEvent:dict,key:str):
        '''
        fix date-parts from the json dict created by crossref e.g.
        "date-parts": [
              [
                1999,
                9,
                5
              ]
            ]
            
        Args:
            rawEvent(dict): the dictionary as parsed by json
            key(str): "start" or "end"

        '''
        if 'date-parts' in rawEvent[key]:
            # get the date-parts to be converted
            dateparts=rawEvent[key]['date-parts']
            # remove the original dict
            rawEvent.pop(key) 
            # get the tuple of values
            datetuple=tuple(dateparts[0])
            # set the year and month default
            year=None
            month=None
            # depending on the date tuples length we get more or less information
            if len(datetuple)==3:
                # a fully specified date
                year,month,day=datetuple
                dt = datetime.datetime(year=year,month=month,day=day)
                date=dt.date()
                rawEvent[f"{key}Date"]=date
            elif len(datetuple)==2:
                # year and month only
                year,month=datetuple
            elif len(datetuple)==1:
                # year only
                year=datetuple[0] 
            else:
                if self.debug:
                    print(f"warning invalid date-tuple {str(datetuple)} found")
            if key=="start":       
                if year is not None: rawEvent["year"]=year
                if month is not None: rawEvent["month"]=month
                       
    
    def jsonFiles(self)->list:  
        '''
        get the list of the json files that have my data
        
        Return:
            list: a list of json file names
        
        '''
        cachePath=self.config.getCachePath()
        self.jsondir=f"{cachePath}/crossref"
        if not os.path.exists(self.jsondir):
                os.makedirs(self.jsondir)
        jsonFiles=sorted(glob.glob(f"{self.jsondir}/crossref-*.json"),key=lambda path:int(re.findall(r'\d+',path)[0]))
        return jsonFiles
    
    def fromJsonFile(self,jsonFilePath):
        '''
        get a single batch of events from the given jsonFilePath
        '''
        eventBatch=None
        with open(jsonFilePath) as jsonFile:
            response=json.load(jsonFile)  
            if 'status' in response:
                if response['status']=='ok':
                    eventBatch=response['message']['items']
        return eventBatch
    
   
    
class CrossrefEventSeriesManager(EventSeriesManager):
    '''
    CrossRef event series handling
    '''
    
    def __init__(self,config:StorageConfig=None):
        '''
        Constructor
        '''
        super().__init__(name="CrossrefEventSeries", clazz=CrossrefEventSeries, tableName="crossef_eventseries",config=config)


    def configure(self):
        '''
        configure me
        '''
        
    def getListOfDicts(self):
        '''
        get my data
        '''
        # TODO Replace this stub
        lod=[{'source':'crossref','eventSeriesId':'dummy'}]
        return lod